# loosely based on: https://github.com/Cahersan/django-schemulator
from collections import OrderedDict
from django import forms


KEYWORDS = {
  #Base keywords
  "label": "title",
  "help_text": "description",
  "initial": "default",
  "required": "required",

  #String type-specific keywords
  "max_length": "maxLength",
  "min_length": "minLength",

  #Numerical type-specific keywords
  "min_value": "minimum",
  "max_value": "maximum",
}

FIELD_TO_TYPE = {
  'IntegerField': 'integer',
  'BooleanField': 'boolean',
  'BooleanField': 'number',
  'TypedChoiceField': ''
}

FIELD_TO_FORMAT = {
  'EmailField': 'email',
  'DateTimeField': 'date-time',
}


def field_to_schema(field):
  field_type = field.__class__.__name__
  schema = {
    'type': FIELD_TO_TYPE.get(field_type, 'string'),
  }

  if not schema['type']:
    #currently only supported for TypedChoiceField
    sample_value = field.coerce('1')
    if isinstance(sample_value, (int, float)):
      schema['type'] = 'integer'
    elif isinstance(sample_value, bool):
      schema['type'] = 'boolean'
    else:
      schema['type'] = 'string'

  if field_type in FIELD_TO_FORMAT:
    schema['format'] = FIELD_TO_FORMAT.get(field_type, None)

  # Setup of JSON Schema keywords
  for (field_attr, schema_attr) in KEYWORDS.items():
    if hasattr(field, field_attr):
      schema[schema_attr] = getattr(field, field_attr)

  # choices needs to be two attrs, so handle it separately
  if hasattr(field, 'choices'):
    optional = not schema.get('required')
    schema['enum'] = [a for a, b in field.choices]
    schema['enumNames'] = [b for a, b in field.choices]
    if not optional and not schema['enum'][0]:
      schema['enum'] = schema['enum'][1:]
      schema['enumNames'] = schema['enumNames'][1:]

  # RJSF doesn't like minLength = null
  if schema.get('minLength', 0) is None:
    schema.pop('minLength')

  if field_type == 'ImageField':
    # RJSF confuses length of file and length of filename, eg "megapixel.png" is 1e6 characters long
    schema.pop('maxLength', None)
    # RJSF is confusde by default None on file field
    schema.pop('default', None)

  for field_attr in ['maxLength', 'title', 'maximum', 'minimum', 'default']:
    if schema.get(field_attr, '') is None:
      schema.pop(field_attr)

  # Set __django_form_field_cls keyword
  schema['__django_form_field_cls'] = field_type
  schema['__widget'] = field.widget.__class__.__name__

  return schema


def form_to_schema(form):
  schema = {
    'type': 'object',
    'properties': OrderedDict([
      (name, field_to_schema(field))
      for (name, field) in form.fields.items()
    ]),
    'required': []
  }

  for name, field in schema['properties'].items():
    if field.pop('required', None):
      schema['required'].append(name)

  if hasattr(form, 'form_title'):
    schema['title'] = form.form_title

  return schema



from django.http import JsonResponse, Http404

from .utils import form_to_schema
import json

FORMS = {}

def FormResponse(form):
    if not form.errors:
        return JsonResponse({})
    return JsonResponse({'errors': form.errors.get_json_data()})


def register(form, form_name=None):
    if isinstance(form, str):
        # register is being used as a decorator and args are curried and reversed
        return lambda actual_form: register(actual_form, form_name=form)
    form_name = form_name or form.__name__
    old_form = FORMS.get(form_name, form)
    if form != old_form:
        e = f"Form with name {form_name} has already been registered.\nOld: {old_form}\nNew:{form}"
        raise ValueError(e)

    FORMS[form_name] = form
    return form


def schema_form(request, form_name, object_id=None, method=None, content_type=None):
    if not form_name in FORMS:
        raise Http404(f"Form with name {form_name} does not exist")

    method = method or request.method
    content_type = content_type or request.headers.get('Content-Type', None)
    form_class = FORMS[form_name]
    _meta  = getattr(form_class, 'Meta', object())
    kwargs = {}
    if object_id and hasattr(_meta, 'model'):
        kwargs['instance'] = _meta.model.objects.get(id=object_id)
    if getattr(_meta, 'login_required', None) and not request.user.is_authenticated:
        return JsonResponse({'error': 'You must be logged in to do this'}, status=403)

    if request.method == "POST":
        if content_type == 'application/json':
            data = json.loads(request.body.decode('utf-8') or "{}")
            form = form_class(data, **kwargs)
        else:
            form = form_class(request.POST, request.FILES, **kwargs)

        form.request = request
        if form.is_valid():
            instance = form.save()
            data = {}
            if instance:
                data = {'id': instance.id, 'name': str(instance)}
            return JsonResponse(data)
        return FormResponse(form)
    schema = form_to_schema(FORMS[form_name]())
    return JsonResponse({'schema': schema})




from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), 'r') as f:
    long_description = f.read()

setup(
  name = 'unrest_schema',
  packages = find_packages(),
  version = '0.1.1',
  description = 'Convert Django forms into json schema',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = 'Chris Cauley',
  author_email = 'chris@lablackey.com',
  url = 'https://github.com/chriscauley/unrest-schema',
  keywords = ['utils','django'],
  install_requires = ['django'],
  license = 'MIT',
  include_package_data = True,
)
